#
# Data validation routines
#

import typing,typing_extensions,types
import functools
import netaddr
import re
from box import Box
from .. import common
from . import get_from_box,set_dots

"""
Common error checking routines:

* wrong_type_text: return the type of the data item, tranforming Box into dict
* wrong_type_message: prints the 'wrong data type' error message
* int_value_error: prints out-of-bounds error message
* check_valid_values: checks scalar or lists for valid values
"""

def wrong_type_text(x : typing.Any) -> str:
  return "dictionary" if isinstance(x,dict) else str(type(x).__name__)

def get_element_path(parent: str, element: typing.Optional[str]) -> str:
  if element:
    return f'{parent}.{element}' if parent else element
  else:
    return parent

wrong_type_help: dict = {}

def wrong_type_message(
      path: str,                                        # Path to the value
      expected: str,                                    # Expected type
      value: typing.Any,                                # Value we got
      key: typing.Optional[str] = None,                 # Optional key within the object
      context: typing.Optional[typing.Any] = None,      # Optional context
      module: typing.Optional[str] = None,              # Module name to display in error messages
                      ) -> None:
  global wrong_type_help

  wrong_type = wrong_type_text(value)
  path = get_element_path(path,key)
  ctxt = f'\n... context: {context}' if context else ''
  expected_help = expected.split(' HELP:')              # Does the error message contain help?
  if len(expected_help) > 1:                            # ... it does, let's see what we can do
    if not expected_help[0] in wrong_type_help:         # Did we print this help before? Adjust context if not
      ctxt = f'\n... FYI: {expected_help[0]} is {expected_help[1]}{ctxt}'
      wrong_type_help[expected_help[0]] = expected_help[1]
    expected = expected_help[0]                         # ... and get the actual target data type name

  wrong_value = 'NWT: ' in expected
  if wrong_value:                                       # String with 'NWT' means 'type is OK, value is incorrect'
    expected = expected.replace('NWT: ','')             # ... but we also don't want NWT in the error message ;)
  else:
    expected += f', found {wrong_type}'                 # A more generic message, add wrong type

  if 'NOATTR:' in path:                                 # Deal with values that are not attributes
    path = path.replace('NOATTR:','')
  else:
    path = f'attribute {path}'

  common.error(
    f'{path} must be {expected}{ctxt}',
    common.IncorrectValue if wrong_value else common.IncorrectType,
    module or 'topology')
  return

def check_valid_values(
      path: str,                                        # Path to the value
      expected: list,                                   # Expected values
      value:    typing.Any,                             # Value we got
      key:      typing.Optional[str] = None,            # Optional key within the object
      context:  typing.Optional[typing.Any] = None,     # Optional context
      module:   typing.Optional[str] = None,            # Module name to display in error messages
      abort:    bool = False,
                      ) -> bool:

  if isinstance(value,list):                            # Deal with lists first
    f = list(filter(lambda x: x not in expected,value)) # ... find all values not in expected values
    if not f:
      return True                                       # ... no unexpected values, cool, get out of here
    value = ','.join(f)                                 # ... otherwise create something to display
  else:
    if value in expected:                               # Not a list? Just check if the value matches one of expected values
      return True

  path = get_element_path(path,key)
  ctxt = f'\n... context: {context}' if context else ''
  common.error(
    f'attribute {path} has invalid value(s): {value}\n... valid values are: {",".join(expected)}{ctxt}',
    common.IncorrectValue,
    module or 'topology')

  if abort:
    raise common.IncorrectValue()
  return False

"""
is_true_int: work around the Python stupidity of bools being ints
"""

def is_true_int(data: typing.Any) -> typing_extensions.TypeGuard[int]:
  return isinstance(data,int) and not isinstance(data,bool)

"""
type_test decorator function -- simplifies the type testing functions

* Get the value (using parent and key)
* Handle empty and boolean values
* Handle valid values and abort-on-error

Required input arguments:

  parent - the parent dictionary of the attribute we want to listify
           (a pointer to the element would be even better, but Python)
  key    - the parent dictionary key
  path   - path of the parent dictionary that would help the user identify
           where the problem is

Optional arguments:

  context       - Additional data identifying the context (usually used for links)
  module        - The caller module (used for error messages)
  valid_values  - list of valid values (applicable to all data types)
  create_empty  - Create an empty element if the value is missing
  true_value    - Replace True with another value
  abort         - Throw an exception after printing an error message

Sample use: make sure the 'config' attribute of a node is list

  must_be_list(node,'config',f'nodes.{node.name}')

"""

def get_value_to_check(
      parent: Box,                                      # Parent object
      key: str,                                         # Key within the parent object, may include dots.
      empty_value: typing.Optional[typing.Any] = None,  # Optional value to use when there is no value
      create_empty: typing.Optional[bool] = None,       # Do we need to create an empty value?
      true_value: typing.Optional[typing.Any] = None,   # Value to use to replace _true_
      false_value: typing.Optional[typing.Any] = None,  # Value to use to replace _false_
      abort: bool = False) -> typing.Any:
  value = get_from_box(parent,key)                      # Otherwise, try to get the value from the parent object
  if value is None:                                     # No value was found, now what?
    if empty_value is not None:                         # ... is there empty value for this data type?
      if create_empty is None:                          # Empty value is defined, and we'll use it to create an empty object if the caller
        create_empty = True                             # did not specify its preferencehs

      if create_empty:                                  # Now for the real deal
        value = empty_value                             # ... if we should create an empty value do so
        set_dots(parent,key.split('.'),empty_value)     # ... and store it in the parent object (dedottifying the key)
      else:
        if abort:                                       # Empty value was specified, 'create_empty' is False, and there's no actual value
          raise common.IncorrectValue()                 # ... raise an exception if requested

    if not key in parent:                               # We can skip the validation of the key is missing.
      return value                                      # ... if the key is there, then we have to make sure the value is valid

  # Handle boolean-to-data-type conversions if the value is bool and the caller specified true_value
  #
  if isinstance(value,bool) and not (true_value is None):
    if value is True:                                   # Replace True with true_value and move on
      value = true_value
      parent[key] = value
    else:
      if false_value is None:                           # If there's no false_value pop the bool option and return None
        parent.pop(key,None)
        return None
      else:
        value = false_value                             # ... otherwise set the false value
        parent[key] = value

  return value

"""
post-validation processing

Validator function could return:

* True -- everything is OK
* False -- failed the validation, error message was already printed
* string -- error message
* callable -- a function returning a replacement value
"""

def post_validation(
      value: typing.Any,
      parent: typing.Optional[Box],
      path: str,
      key: typing.Optional[str],
      expected: typing.Any,
      context: typing.Optional[str] = None,
      module: typing.Optional[str] = None,
      abort: bool = False,
      test_function: typing.Any = None) -> typing.Any:

  if isinstance(expected,(bool,str)):
    if not expected is True:
      if isinstance(expected,str):
        wrong_type_message(
          path=path,
          key=None if parent is None else key,
          expected=expected,
          value=value,context=context,
          module=module)
      if abort:
        raise common.IncorrectType()
      return None
  elif isinstance(expected,types.FunctionType):
    value = expected(value)
    if not parent is None:
      parent[key] = value
  else:
    common.fatal(f'Validator function {test_function} returned unexpected value {expected}')

  return value

def type_test(
      false_value: typing.Optional[typing.Any] = None,
      empty_value: typing.Optional[typing.Any] = None) -> typing.Callable:

  def test_wrapper(test_function: typing.Callable) -> typing.Callable:

    # Generic data type validation framework
    #
    # Use it as a decorator wrapper around the actual testing function
    #
    @functools.wraps(test_function)
    def execute_test(
          parent: typing.Optional[Box],                     # Parent object
          key: str,                                         # Key within the parent object, may include dots.
          path: str,                                        # Path to parent object, used in error messages
          context: typing.Optional[typing.Any] = None,      # Additional context (use when verifying link values)
          module: typing.Optional[str] = None,              # Module name to display in error messages
          valid_values: typing.Optional[list] = None,       # List of valid values
          create_empty: typing.Optional[bool] = None,       # Do we need to create an empty value?
          true_value: typing.Optional[typing.Any] = None,   # Value to use to replace _true_ (false_values used to replace _false_)
          abort: bool = False,                              # Abort on error
          **kwargs: typing.Any) -> typing.Optional[typing.Any]:

      if parent is None:                                    # No parent => key is the value to check
        value = key
      else:
        value = get_value_to_check(
                  parent=parent,
                  key=key,
                  empty_value=empty_value,
                  create_empty=create_empty,
                  true_value=true_value,
                  false_value=false_value,
                  abort=abort)
        if not key in parent:
          return value

      expected = test_function(value,**kwargs)              # Now call the validator function with the item value
      value = post_validation(
                value=value,
                parent=parent,
                path=path,
                key=key,
                expected=expected,
                context=context,
                module=module,
                abort=abort,
                test_function=test_function)

      # Finally, check valid values (if specified)
      #
      if valid_values:
        check_valid_values(
          path=path,
          key=None if parent is None else key,
          value=value,
          expected=valid_values,
          context=context,
          module=module,
          abort=abort)

      # And return whatever the final value is (considering empty, true, and transformed values)
      #
      return value

    return execute_test

  return test_wrapper

"""
Individual data type validators
===============================

Most validators check the instance type and return a string error message

Exceptions:

* List validator returns a transformation function when a scalar could be converted to a list
* Integer validation can include minimum/maximum values
"""

@type_test(false_value=[],empty_value=[])
def must_be_list(value: typing.Any) -> typing.Union[bool,str,typing.Callable]:

  def transform_to_list(value: typing.Any) -> list:
    return [ value ]

  if isinstance(value,(str,int,float,bool)):            # Handle scalar-to-list transformations with a callback function
    return transform_to_list

  return True if isinstance(value,list) else 'a scalar or a list'

@type_test(false_value={},empty_value={})
def must_be_dict(value: typing.Any) -> typing.Union[bool,str,typing.Callable]:
  return True if isinstance(value,dict) else 'a dictionary'

@type_test(false_value='')
def must_be_string(value: typing.Any) -> typing.Union[bool,str]:
  return True if isinstance(value,str) else 'a string'

@type_test(false_value='')
def must_be_str(value: typing.Any) -> typing.Union[bool,str]:
  return True if isinstance(value,str) else 'a string'

@type_test()
def must_be_id(value: typing.Any) -> typing.Union[bool,str]:
  if not isinstance(value,str) or not re.fullmatch('[a-zA-Z_][a-zA-Z0-9_-]{0,15}',value):
    return 'an identifier HELP:a string containing up to 16 alphanumeric characters, numbers and underscores'

  return True

@type_test()
def must_be_int(
      value: typing.Any,
      min_value:  typing.Optional[int] = None,          # Minimum value
      max_value:  typing.Optional[int] = None,          # Maximum value
                ) -> typing.Union[bool,str,typing.Callable]:

  def transform_to_int(value: typing.Any) -> int:
    return int(value)

  if isinstance(value,str):                             # Try to convert STR to INT
    try:
      transform_to_int(value)
      return transform_to_int
    except:
      pass

  if not isinstance(value,int):                         # value must be an int
    return 'an integer'

  if isinstance(value,bool):                            # but not a bool
    return 'NWT: a true integer (not a bool)'

  if isinstance(min_value,int) and isinstance(max_value,int):
    if value < min_value or value > max_value:
      return f'NWT: an integer between {min_value} and {max_value}'
  elif isinstance(min_value,int):
    if value < min_value:
      return f'NWT: an integer larger or equal to {min_value}'
  elif isinstance(max_value,int):
    if value > max_value:
      return f'NWT: an integer less than or equal to {max_value}'

  return True

@type_test()
def must_be_bool(value: typing.Any) -> typing.Union[bool,str,typing.Callable]:

  def transform_to_bool(value: typing.Any) -> bool:
    if value == 'True' or value == 'true':
      return True

    if value == 'False' or value == 'false':
      return False 

    raise ValueError('invalid boolean literal -- use true or false')

  if isinstance(value,str):                             # Try to convert STR to INT
    try:
      transform_to_bool(value)
      return transform_to_bool
    except:
      pass

  return True if isinstance(value,bool) else 'a boolean'

@type_test()
def must_be_asn(value: typing.Any) -> typing.Union[bool,str]:
  err = 'an AS number (integer between 1 and 65535)'
  if not isinstance(value,int) or isinstance(value,bool):             # value must be an int
    return err

  if value < 0 or value > 65535:
    return 'NWT: '+err

  return True

#
# Testing for IPv4 and IPv6 addresses is nasty, as netaddr module happily mixes IPv4 and IPv6
#
@type_test()
def must_be_ipv4(value: typing.Any, use: str) -> typing.Union[bool,str]:
  if isinstance(value,bool):                                          # bool values are valid only on interfaces
    if use not in ('interface','prefix'):
      return 'NWT: an IPv4 address (boolean value is valid only on an interface)'
    else:
      return True

  if isinstance(value,int):                                           # integer values are valid only as IDs (OSPF area)
    if use not in ('id','interface'):
      return 'NWT: an IPv4 prefix (integer value is only valid as a 32-bit ID)'
    if value < 0 or value > 2**32-1:
      return 'NWT: an IPv4 address or an integer between 0 and 2**32'
    return True

  if not isinstance(value,str):
    return 'IPv4 prefix' if use == 'prefix' else 'IPv4 address'

  if '/' in value:
    if use == 'id':                                                   # IDs should not have a prefix
      return 'NWT: IPv4 address (not prefix)'
  else:
    if use == 'prefix':                                               # prefix must have a /
      return 'NWT: IPv4 prefix'

  try:
    parse = netaddr.IPNetwork(value)                                  # now let's check if we have a valid address
  except Exception as ex:
    return "NWT: IPv4 " + ("address or " if use != 'prefix' else "") + "prefix"

  try:                                                                # ... and finally we have to check it's a true IPv4 address
    parse.ipv4()
    if parse.is_ipv4_mapped():
      return "NWT: IP address in IPv4 format"
  except Exception as ex:
    return "NWT: IPv4 address/prefix"

  return True

@type_test()
def must_be_ipv6(value: typing.Any, use: str) -> typing.Union[bool,str]:
  if isinstance(value,bool):                                          # bool values are valid only on interfaces
    if use not in ('interface','prefix'):
      return 'NWT: an IPv6 address (boolean value is valid only on an interface)'
    else:
      return True

  if isinstance(value,int):                                           # integer values are valid only as IDs (OSPF area)
    if use not in ('interface'):
      return 'NWT: an IPv6 prefix (integer value is only valid as an inteface offset)'
    return True

  if not isinstance(value,str):
    return 'IPv6 prefix' if use == 'prefix' else 'IPv6 address'

  if not '/' in value:
    if use == 'prefix':                                               # prefix must have a /
      return 'NWT: IPv6 prefix (not an address)'

  try:
    parse = netaddr.IPNetwork(value)                                  # now let's check if we have a valid address
  except Exception as ex:
    return "NWT: IPv6 " + ("address or " if use != 'prefix' else "") + "prefix"

  if parse.is_ipv4_mapped():                                          # This is really an IPv4 address, but it looks like IPv6, so OK
    return True

  try:                                                                # ... and finally we have to check it's a true IPv6 address
    parse.ipv4()
    return "NWT: IPv6 (not an IPv4) address"                          # If we could get IPv4 address out of it, it clearly is not
  except Exception as ex:
    pass

  return True

@type_test()
def must_be_mac(value: typing.Any) -> typing.Union[bool,str]:
  if not isinstance(value,str):
    return 'MAC address'

  try:
    parse = netaddr.EUI(value)                                        # now let's check if we have a MAC address
  except Exception as ex:
    return "MAC address"

  return True

@type_test()
def must_be_net(value: typing.Any) -> typing.Union[bool,str]:
  if not isinstance(value,str):
    return 'IS-IS NET/NSAP'

  if not re.fullmatch('[0-9a-f.]+',value):
    return 'NWT: IS-IS NET/NSAP containing hexadecimal digits or dots'

  if len(value.replace('.','')) % 2 != 0:
    return 'NWT: IS-IS NET/NSAP containing even number of hexadecimal digits'

  return True

@type_test()
def must_be_rd(value: typing.Any) -> typing.Union[bool,str]:
  if isinstance(value,int) or value is None:                          # Accept RD/RT offets and trust the modules to do the right thing
    return True                                                       # Also: RD set to None can be used to prevent global-to-node RD inheritance

  if not isinstance(value,str):                                       # Otherwise it must be a string
    return "route distinguisher"

  try:
    (rdt,rdi) = value.split(':')
  except Exception as ex:
    return "NWT: route distinguisher in asn:id or ip:id format"

  try:
    rdi_parsed = int(rdi)
  except Exception as ex:
    return "NWT: an RD in asn:id or ip:id format where id is an integer value"

  try:
    rdt_parsed = int(rdt)
  except Exception as ex:
    try:
      if '/' in rdt:
        return "route distinguisher in asn:id or ip:id format"
      netaddr.IPNetwork(rdt)
    except Exception as ex:
      return "NWT: an RD in asn:id or ip:id format where asn is an integer or ip is an IPv4 address"

  return True



import typing

from . import _Module,_routing,_dataplane
from box import Box
from .. import common
from .. import data
from ..data.validate import must_be_int,must_be_dict,must_be_string
from ..augment import devices

"""
validate_evpn_list

* If missing, create evpn.vlans list with all VLANs with 'vni' attribute
* Validate that the evpn.vlans list is a list with valid local or global VLAN names
* If missing, create evpn.vrfs list with all VRFs with 'evpn.transit_vni' attribute
* Validate that the evpn.vrfs list is a list with valid local or global VRF names

The heavy lifting is done in a shared _dataplane function
"""
def validate_evpn_lists(toponode: Box, obj_path: str, topology: Box, create: bool = False) -> None:
  _dataplane.validate_object_reference_list(
    parent=toponode if not toponode is topology else None,
    parent_path=obj_path,
    topology=topology, 
    list_name='evpn.vlans',
    reference_dictionary='vlans',
    reference_name='VLAN',
    create_default=create,
    merge_topology=False,
    default_filter=lambda v: 'vni' in v and v.vni,
    module='evpn')

  _dataplane.validate_object_reference_list(
    parent=toponode if not toponode is topology else None,
    parent_path=obj_path,
    topology=topology, 
    list_name='evpn.vrfs',
    reference_dictionary='vrfs',
    reference_name='VLAN',
    create_default=create,
    merge_topology=False,
    default_filter=lambda v: False if not isinstance(v,Box) else data.get_from_box(v,'evpn.transit_vni'),
    module='evpn')

def enable_evpn_af(node: Box, topology: Box) -> None:
  bgp_session = data.get_from_box(node,'evpn.session') or []

  # Enable EVPN AF on all BGP neighbors with the correct session type
  # that also use EVPN module
  #
  for bn in node.bgp.get('neighbors',[]):
    if bn.type in bgp_session and 'evpn' in topology.nodes[bn.name].get('module'):
      bn.evpn = True

def get_usable_evpn_asn(topology: Box) -> int:
  asn = data.get_from_box(topology,'evpn.as') or \
        data.get_from_box(topology,'vrf.as') or \
        data.get_from_box(topology,'bgp.as')

  if asn and data.is_true_int(asn):
    return asn

  common.error(
    f'Cannot get a usable global AS number to use in EVPN route targets',
    common.IncorrectValue,
    'evpn')
  return 0

def vlan_based_service(vlan: Box, vname: str, topology: Box) -> None:
  evpn  = vlan.evpn
  epath = f'vlans.{vname}.evpn'
  evpn.evi = evpn.evi or vlan.id                                    # Default EVI value: VLAN ID
  asn = get_usable_evpn_asn(topology)
  must_be_int(
    evpn,'evi',epath,
    module='evpn',
    min_value=1,max_value=65535)                                    # Check EVI data type in range
  for rt in ('import','export'):                                    # Default RT value
    if not rt in evpn:                                              # ... BGP ASN:vlan ID
      evpn[rt] = [ f"{asn}:{evpn.evi}" ]

def vlan_aware_bundle_service(vlan: Box, vname: str, topology: Box) -> None:
  vrf_name = vlan.vrf
  if not vrf_name in topology.vrfs:
    common.error(
      f'VXLAN-enabled VLAN {vname} that is part of VLAN bundle must belong to a global VRF',
      common.IncorrectValue,
      'evpn')
    return

  if 'evpn' in vlan:                                                # VLAN that is part of a bundle cannot have EVI/RT/RD attributes
    common.error(
      f'VLAN {vname} is part of a VLAN bundle {vrf_name} and cannot have EVPN-related attributes',
      common.IncorrectValue,
      'evpn')
    return

  vrf = topology.vrfs[vrf_name]
  must_be_dict(
    parent=vrf,
    key='evpn',
    path=f'vrfs.{vrf_name}',
    create_empty=True)
 
  evpn = vrf.evpn
  vlan.evpn.bundle = evpn.bundle                                    # Copy the bundle type into VLAN
  if not 'evi' in evpn:                                             # If needed, set EVI attribute for the global VRF
    evpn.evi = vrf.id                                               # ... to the VRF ID

  vlan.evpn.evi = evpn.evi
  for k in ('vlans','vlan_ids'):
    if not k in evpn:                                               # Is this the first EVPN-enabled VLAN in this VRF?
      evpn[k] = []                                                  # ... create an empty list of VLANs
  evpn.vlans.append(vname)                                          # Finally, add VLAN name to the list of MAC VRF VLANs
  evpn.vlan_ids.append(topology.vlans[vname].id)                    # ... and a VLAN ID to list of EVPN-enabled VLAN tags


"""
Figure out whether we have to treat a VLAN as part of a VLAN bundle or an independent VLAN service
"""

def get_vlan_bundle_flag(vlan: Box, vname: str, topology: Box) -> bool:
  if not 'vrf' in vlan:                                             # VLAN not part of a VRF ==> VLAN service
    return False

  vrf = topology.vrfs[vlan.vrf]
  if not data.get_from_box(vrf,'evpn.bundle'):                      # VRF does not have evpn.bundle attribute ==> VLAN service
    return False

  must_be_string(
    parent=vrf,
    key='evpn.bundle',
    path=f'vrfs.{vlan.vrf}',
    module='evpn',
    valid_values=topology.defaults.evpn.attributes.bundle)          # Validate evpn.bundle value

  return True

"""
Create VLAN services -- either as independent VLANs or as a VRF-based bundle
"""

def create_vlan_service(vname: str, topology: Box) -> None:
  vlan = topology.vlans[vname]
  if not get_vlan_bundle_flag(vlan,vname,topology):
    vlan_based_service(vlan,vname,topology)
  else:
    vlan_aware_bundle_service(vlan,vname,topology)

"""
Validate that there's no evpn.bundle setting on node VRFs
"""

def validate_no_node_vrf_bundle(node: Box, topology: Box) -> None:
  if not 'vrfs' in node:                                            # No VRFs defined in the node, move on
    return
  for vname,vdata in node.vrfs.items():
    if not data.get_from_box(vdata,'evpn.bundle'):                  # EVPN bundle not set in VRF, move on
      continue
    common.error(
      f'VRF {vname} in node {node.name} cannot have evpn.bundle attribute',
      common.IncorrectValue,
      'evpn',
      hint='node_bundle')

"""
Validate transit VNI values and register them with the VNI set
"""
def register_static_transit_vni(topology: Box) -> None:
  vni_set = _dataplane.get_id_set('vni')
  for vrf_name,vrf_data in topology.get('vrfs',{}).items():
    must_be_dict(vrf_data,'evpn',f'vrfs.{vrf_name}',create_empty=False)

    transit_vni = data.get_from_box(vrf_data,'evpn.transit_vni')
    if data.is_true_int(transit_vni):
      if transit_vni in vni_set:
        common.error(
          f'transit VNI {transit_vni} for VRF {vrf_name} is already used elsewhere',
          common.IncorrectValue,
          'evpn')
        continue
      vni_set.add(transit_vni)

  for n in topology.nodes.values():
    if not 'vrfs' in n:
      continue

    for vrf_name,vrf_data in n.vrfs.items():
      if data.get_from_box(vrf_data,'evpn.transit_vni'):
        common.error(
          f'evpn.transit_vni can be specified only on global VRFs (found in {vrf_name} on {n.name}',
          common.IncorrectValue,
          'evpn')

"""
Set transit VNI values for symmetrical IRB VRFs (REFACTOR to use _dataplane)
"""
def get_next_vni(start_vni: int, used_vni_list: typing.List[int]) -> int:
  while True:
    start_vni = start_vni + 1
    if not start_vni in used_vni_list:
      return start_vni

def vrf_transit_vni(topology: Box) -> None:
  if not 'vrfs' in topology:
    return

  vni_list: typing.List[int] = []                               # List of static transit VNIs
  vni_error = False                                             # "A horrible error" flag that causes abort after the first loop
  vni_count = 0                                                 # Number of VRFs with evpn.transit_vni
  evpn_transport = data.get_from_box(topology,'evpn.transport') or 'vxlan'

  for vrf_name,vrf_data in topology.vrfs.items():               # First pass: build a list of statically configured VNIs
    vni = data.get_from_box(vrf_data,'evpn.transit_vni')        # transit_vni makes no sense with MPLS transport
    if vni and evpn_transport != 'vxlan':
      common.error(
        f'evpn.transit_vni in VRF {vrf_name} is not allowed with mpls evpn.transport',
        common.IncorrectValue,
        'evpn')
      vni_error = True
      continue

    if not vni is None:
      vni_count = vni_count + 1                                 # Count number of VRFs with evpn.transit_vni attribute
    if not data.is_true_int(vni):                               # Skip non-integer values, no need to check them at this time
      continue
    if vni in vni_list:
      common.error(
        f'VRF {vrf_name} is using the same EVPN transit VNI as another VRF',
        common.IncorrectValue,
        'evpn')
      continue  
    vni_list.append( vni )                                      # Insert it to detect duplicates elsewhere

  if vni_error:                                                 # Found serious errors, makes no sense to continue
    return
  if not vni_count:                                             # No VRF found with evpn.transit_vni, no need to waste further CPU cycles
    return

  vni_start = topology.defaults.evpn.start_transit_vni
  for vrf_name,vrf_data in topology.vrfs.items():               # Second pass: set transit VNI values for VRFs with "transit_vni: True"
    if isinstance(data.get_from_box(vrf_data,'evpn.transit_vni'),str):
      continue                                                  # Skip transit_vni string values (will be checked in third pass)
    transit_vni = must_be_int(
                    vrf_data,
                    key='evpn.transit_vni',
                    path=f'vrfs.{vrf_name}',
                    module='evpn',
                    min_value=4096,                             # As recommended by Cisco, outside of VLAN range
                    max_value=16777215,
                    true_value=vni_start)                       # Make sure evpn.transit_vni is an integer
    if transit_vni == vni_start:                                # If we had to assign the default value, increment the default transit VNI
      vni_start = get_next_vni(vni_start,vni_list)

  for vrf_name,vrf_data in topology.vrfs.items():               # Third pass: set shared VNI values across VRFs
    if vrf_data is None:                                        # Skip empty VRF definitions
      continue
    transit_vni = data.get_from_box(vrf_data,'evpn.transit_vni')
    if not isinstance(transit_vni,str):                         # Skip if transit_vni is not a string
      continue
    if not transit_vni in topology.vrfs:                        # Does transit VNI refer to a valid VRF name?
      common.error(
        f'evpn.transit_vni "{transit_vni}" in VRF {vrf_name} does not refer to a valid VRF',
        common.IncorrectValue,
        'evpn')
      continue
    foreign_vni = data.get_from_box(topology.vrfs,f'{transit_vni}.evpn.transit_vni')
    if not data.is_true_int(foreign_vni):
      common.error(
        f'evpn.transit_vni "{transit_vni}" in VRF {vrf_name} refers to a VRF that does not have a valid evpn.transit_vni',
        common.IncorrectValue,
        'evpn')
      continue
    vrf_data.evpn.transit_vni = foreign_vni

def vrf_irb_setup(topology: Box) -> None:
  for vrf_name,vrf_data in topology.get('vrfs',{}).items():
    if not 'evpn' in vrf_data:                                  # VRF without EVPN data is definitely not doing IRB
      continue
    
    if not 'evi' in vrf_data.evpn:
      vrf_data.evpn.evi = vrf_data.id

"""
Copy global evpn.vlans, evpn.vrfs, and evpn.transport into node evpn data structures.
"""
def copy_global_evpn_lists(node: Box, topology: Box) -> None:
  g_evpn = topology.get('evpn',{})
  for k in ('vlans','vrfs','transport'):
    if k in g_evpn and not k in node.evpn:
      node.evpn[k] = g_evpn[k]

"""
Trim node evpn.vlans and evpn.vrfs lists to refer only to on-node VLANs and VRFs
"""
def trim_node_evpn_lists(node: Box) -> None:
  if not 'evpn' in node:                                        # No on-node EVPN structures, get out of here
    return

  for ref_list in ('vlans','vrfs'):
    if ref_list in node.evpn:                                   # Filter EVPN list to refer only to on-node objects
      node.evpn[ref_list] = [ obj_name for obj_name in node.get(ref_list,{}).keys() if obj_name in node.evpn[ref_list] ]

"""
Set local evpn.rd for evpn-enabled VLANs and VRFs
"""
def set_local_evpn_rd(node: Box) -> None:
  for obj in ('vlans','vrfs'):
    if not obj in node:
      continue

    for o_name,o_data in node[obj].items():                     # Iterate over VLANs/VRFs
      if not 'evpn' in o_data:
        continue                                                # Skip non-EVPN objects
      if not 'evi' in o_data.evpn:
        common.error(
          f'{obj}.{o_name} on node {node.name} has EVPN attributes but no evpn.evi ({o_data.evpn})',
          common.IncorrectValue,
          'evpn')
        continue
      #
      # Set local RD unless a static RD was set
      if not 'rd' in o_data.evpn:
        o_data.evpn.rd = f'{node.bgp.router_id}:{o_data.evpn.evi}'

"""
Check whether all VLANs needed for an asymmetric IRB VRF are present on a node
"""

def check_asym_vlan(vrf_name: str, node: Box, topology: Box) -> None:
  if not 'vlans' in topology:                                   # No global VLANs, nothing to check
    return

  for vl_name,vl_data in topology.vlans.items():                # Iterate over global VLANs
    if not vrf_name in vl_data.get('vrf',''):                   # The VLAN is not in current VRF, skip it
      continue

    if not 'evpn' in vl_data:                                   # VLAN is not EVPN-enabled, so it doesn't have to be present everywhere
      continue

    if vl_name in node.get('vlans',{}):                         # Is the VLAN present on the node?
      continue                                                  # Yeah, everything OK

    common.error(
      f'VLAN {vl_name} -- part of VRF {vrf_name} that uses asymmetric IRB -- is not present on node {node.name}',
      common.IncorrectValue,
      module='evpn',
      hint='irb_group')

"""
Check whether VXLAN IRB mode is supported by the device
"""
def check_node_vrf_irb(node: Box, topology: Box) -> None:
  features = devices.get_device_features(node,topology.defaults)
  evpn_transport = data.get_from_box(node,'evpn.transport') or 'vxlan'

  for vrf_name,vrf_data in node.get('vrfs',{}).items():
    if not vrf_data.get('af',None):                             # Cannot do IRB without L3 addresses ;)
      continue

    symmetric_irb = data.get_from_box(vrf_data,'evpn.transit_vni') or evpn_transport == 'mpls'
    if symmetric_irb:
      if not features.evpn.irb and evpn_transport == 'vxlan':   # ... does this device support IRB?
        common.error(
          f'VRF {vrf_name} on {node.name} uses symmetrical EVPN IRB which is not supported by {node.device} device',
          common.IncorrectValue,
          'evpn')
        continue

      vrf_data.pop('ospf',None)                                 # Remove OSPF from EVPN IRB VRF
    else:
      if not features.evpn.asymmetrical_irb:                    # ... does this device asymmetrical IRB -- is it supported?
        common.error(
          f'VRF {vrf_name} on {node.name} uses asymmetrical EVPN IRB which is not supported by {node.device} device',
          common.IncorrectValue,
          'evpn')
        continue

      check_asym_vlan(vrf_name,node,topology)

"""
Check whether the node supports the requested EVPN bundle type
"""

def check_node_vrf_bundle(node: Box, topology: Box) -> None:
  features = devices.get_device_features(node,topology.defaults)      # Get device features
  for vrf_name,vrf_data in node.get('vrfs',{}).items():               # Iterate over all VRFs defined on the device
    b_type = data.get_from_box(vrf_data,'evpn.bundle')                # Get evpn.bundle value, skip if not defined
    if not b_type:
      continue
    if not b_type in features.evpn.bundle:                            # EVPN bundle type not supported by the device
      common.error(
        f"'{b_type}'' EVPN bundle service used in VRF {vrf_name} is not supported by device {node.device} (node: {node.name})",
        common.IncorrectValue,
        'evpn')

class EVPN(_Module):

  def module_init(self, topology: Box) -> None:
    topology.defaults.vxlan.flooding = 'evpn'

  def module_pre_transform(self, topology: Box) -> None:
    register_static_transit_vni(topology)
    must_be_string(
      parent=topology,
      key='evpn.transport',
      path='topology',
      valid_values=['vxlan','mpls'],
      module='evpn')

  def node_pre_transform(self, node: Box, topology: Box) -> None:
    validate_no_node_vrf_bundle(node,topology)

  def module_post_node_transform(self, topology: Box) -> None:
    validate_evpn_lists(topology,'topology',topology,create=True)
    for n in topology.nodes.values():
      validate_evpn_lists(n,f'nodes.{n.name}',topology,create=False)

    vrf_transit_vni(topology)
    for vname in data.get_from_box(topology,'evpn.vlans') or []:
      create_vlan_service(vname,topology)

    vrf_irb_setup(topology)

  """
  Node post-transform: runs after VXLAN module

  * Enable EVPN address family
  * Build node-specific evpn.vlans and epvn.vrfs lists
  * Calculate final RD values

  """
  def node_post_transform(self, node: Box, topology: Box) -> None:
    enable_evpn_af(node,topology)
    _routing.router_id(node,'bgp',topology.pools)                 # Make sure we have a usable router ID
    copy_global_evpn_lists(node,topology)
    check_node_vrf_irb(node,topology)
    check_node_vrf_bundle(node,topology)
    trim_node_evpn_lists(node)
    set_local_evpn_rd(node)



#
# External provider module
#
import subprocess
import typing
from box import Box

from . import _Provider
from .. import common

class External(_Provider):
  def augment_node_data(self, node: Box, topology: Box) -> None:
    common.print_verbose('Augmenting node data for External')
    # Cleanup MGMT MAC Address (since it's useless for us)
    node.mgmt.pop('mac',None)

  def pre_start_lab(self, topology: Box) -> None:
    common.print_verbose('pre-start hook for External')
    print('*** Please make sure your physical topology reflects the following cabling:')
    print('')
    with open('external.txt', 'r') as f:
      print(f.read())
    print('')
    print('*** Please make sure your physical topology reflects the above cabling.')
    if input('Do you want to continue [Y/n]: ') != 'Y':
      common.fatal('Aborting...')



#
# Create Ansible inventory
#
import typing

import yaml
import os
from box import Box

from .. import common
from . import _TopologyOutput,check_writeable
from ..augment import nodes
from ..augment import devices

forwarded_port_name = { 'ssh': 'ansible_port', }

def copy_provider_inventory(host: Box, p_data: Box) -> None:
  if 'inventory' in p_data:
    for k,v in p_data.inventory.items():
      host[k] = v

  if 'inventory_port_map' in p_data and 'forwarded' in p_data:
    for k,v in p_data.inventory_port_map.items():
      if k in p_data.forwarded:
        host[v] = p_data.forwarded[k] + host.id

def copy_device_provider_group_vars(host: Box, node: Box, topology: Box) -> None:
  p_data = devices.get_provider_data(node,topology.defaults)
  if not 'group_vars' in p_data:
    return

  for k,v in p_data.group_vars.items():
    if not k in host:
      host[k] = v

def provider_inventory_settings(host: Box, node: Box, topology: Box) -> None:
  defaults = topology.defaults
  node_provider = devices.get_provider(node,topology)
  p_data = defaults.providers[node_provider]
  if p_data:
    copy_provider_inventory(host,p_data)

  if 'provider' in node:                                              # Is the node using a secondary provider?
    copy_device_provider_group_vars(host,node,topology)

topo_to_host = { 'mgmt.ipv4': 'ansible_host', 'hostname': 'ansible_host', 'id': 'id' }
topo_to_host_skip = [ 'name','device' ]

def ansible_inventory_host(node: Box, topology: Box) -> Box:
  defaults = topology.defaults
  host = Box({})
  for (node_key,inv_key) in topo_to_host.items():
    if "." in node_key:
      value = node[node_key]
    else:
      value = node.get(node_key,None)
    if value:
      host[inv_key] = value

  for (k,v) in node.items():
    if not k in topo_to_host_skip:
      host[k] = v

  provider_inventory_settings(host,node,topology)
  return host

def create(topology: Box) -> Box:
  inventory = Box({},default_box=True,box_dots=True)

  inventory.all.vars.netlab_provider = topology.defaults.provider
  inventory.all.vars.netlab_name = topology.name

  inventory.modules.hosts = {}
  inventory.custom_configs.hosts = {}

  defaults = topology.defaults

  if 'module' in topology:
    inventory.modules.vars.netlab_module = topology.module

  if 'addressing' in topology:
    inventory.all.vars.pools = topology.addressing
    for name,pool in inventory.all.vars.pools.items():
      for k in list(pool.keys()):
        if ('_pfx' in k) or ('_eui' in k):
          del pool[k]

  for name,node in topology.nodes.items():
    group = node.get('device','all')
    inventory[group].hosts[name] = ansible_inventory_host(node,topology)

    if 'module' in node:
      inventory.modules.hosts[name] = {}

    if 'config' in node:
      inventory.custom_configs.hosts[name] = {}

  if 'devices' in defaults:
    for group in inventory.keys():
      if group in defaults.devices:
        devdata = defaults.devices[group]
        group_vars = devdata.group_vars + devdata[defaults.provider].group_vars
        if group_vars:
          inventory[group]['vars'] = group_vars

  if not 'groups' in topology:
    return inventory

  for gname,gdata in topology.groups.items():
    if not gname in inventory:
      inventory[gname] = { 'hosts': {} }

    if 'vars' in gdata:
      inventory[gname].vars = inventory[gname].get('vars',{}) + gdata.vars

    if 'members' in gdata:
      for m in gdata.members:
        if m in topology.nodes:
          if not m in inventory[gname].hosts:
            inventory[gname].hosts[m] = {}
        elif m in topology.groups:
          inventory[gname].children[m] = {}

  return inventory

# Note to self: the dump function is used by the testing scripts. Do not remove
#
def dump(data: Box) -> None:
  print("Ansible inventory data")
  print("===============================")
  inventory = create(data)
  print(common.get_yaml_string(inventory))

def write_yaml(data: Box, fname: str, header: str) -> None:
  dirname = os.path.dirname(fname)
  if dirname and not os.path.exists(dirname):
    os.makedirs(dirname)

  with open(fname,"w") as output:
    output.write(header)
    output.write(common.get_yaml_string(data))
    output.close()

min_inventory_data = [ 'id','ansible_host','ansible_port','ansible_connection','ansible_user','ansible_ssh_pass' ]

def ansible_inventory(topology: Box, fname: typing.Optional[str] = 'hosts.yml', hostvars: typing.Optional[str] = 'dirs') -> None:
  inventory = create(topology)

#  import ipdb; ipdb.set_trace()
  header = "# Ansible inventory created from %s\n#\n" % topology.get('input','<unknown>')

  if not fname:
    fname = 'hosts.yml'
  if not hostvars:
    hostvars = "dirs"

  if hostvars == "min":
    write_yaml(inventory,fname,header)
    print("Created single-file Ansible inventory %s" % fname)
    return

  for g in inventory.keys():
    gvars = inventory[g].pop('vars',None)
    if gvars:
      write_yaml(gvars,'group_vars/%s/topology.yml' % g,header)
      if not common.QUIET:
        print("Created group_vars for %s" % g)

    if 'hosts' in inventory[g]:
      hosts = inventory[g]['hosts']
      for h in hosts.keys():
        if not hosts[h]:
          continue
        min_host = Box({})
        vars_host = Box({})
        for item in hosts[h].keys():
          if item in min_inventory_data:
            min_host[item] = hosts[h][item]
          else:
            vars_host[item] = hosts[h][item]

        write_yaml(vars_host,'host_vars/%s/topology.yml' % h,header)
        if not common.QUIET:
          print("Created host_vars for %s" % h)
        hosts[h] = min_host

  write_yaml(inventory,fname,header)
  print("Created minimized Ansible inventory %s" % fname)

def ansible_config(config_file: typing.Union[str,None] = 'ansible.cfg', inventory_file: typing.Union[str,None] = 'hosts.yml') -> None:
  if not config_file:
    config_file = 'ansible.cfg'
  if not inventory_file:
    inventory_file = 'hosts.yml'

  with open(config_file,"w") as output:
    output.write(common.template('ansible.cfg.j2',{ 'inventory': inventory_file or 'hosts.yml' },'templates','ansible'))
    output.close()
    if not common.QUIET:
      print("Created Ansible configuration file: %s" % config_file)

class AnsibleInventory(_TopologyOutput):

  def write(self, topology: Box) -> None:
    check_writeable('Ansible inventory')
    hostfile = self.settings.hostfile or 'hosts.yml'
    configfile = self.settings.configfile or 'ansible.cfg'
    output_format = None

    if hasattr(self,'filenames'):
      hostfile = self.filenames[0]
      if len(self.filenames) > 1:
        configfile = self.filenames[1]
      if len(self.filenames) > 2:
        common.error('Extra output filename(s) ignored: %s' % str(self.filenames[2:]),common.IncorrectValue,'ansible')

    if self.format:
      output_format = self.format[0]
    
    # Creates a "ghost clean" topology
    # (AKA, remove unmanaged devices)
    ansible_topology = nodes.ghost_buster(topology)

    ansible_inventory(ansible_topology,hostfile,output_format)
    ansible_config(configfile,hostfile)



'''
Add group support. This module handles group-related data structures and
data transformation. The end result is a dictionary of groups in 'groups'
top-level element. That dictionary is merged from global- and node-level parameters
'''

import typing
import re

from box import Box

from .. import common
from .. import data
from .. import modules
from ..modules import bgp
from ..data.validate import validate_attributes
from ..data.types import must_be_dict,must_be_list,must_be_string,must_be_id
from . import nodes

'''
Return members of the specified group. Recurse through child groups if needed
'''
def group_members(topology: Box, group: str, count: int = 0) -> list:
  members: typing.List[str] = []
  if not group in topology.groups:  # pragma: no cover (just-in case catch, impossible to get here)
    common.error(
      f'Internal error: unknown group {group}',
      common.IncorrectValue,
      'groups')
    return []

  if count > 99:                    # pragma: no cover (impossible to get here, recursive groups are checked elsewhere)
    common.fatal(
      'Recursive group definition, aborting',
      'groups')

  for m in topology.groups[group].members:
    if m in topology.nodes:
      members = members + [ m ]
    if m in topology.groups:
      members = members + group_members(topology,m,count + 1)

  return members

'''
Check validity of 'groups' data structure
'''
def check_group_data_structure(topology: Box) -> None:
  if not 'groups' in topology:
    topology.groups = Box({},default_box=True,box_dots=True)

  if must_be_dict(topology,'groups','topology',create_empty=True,module='groups') is None:
    return

  '''
  Transform group-as-list into group-as-dictionary
  '''
  for grp in topology.groups.keys():
    if isinstance(topology.groups[grp],list):
      topology.groups[grp] = { 'members': topology.groups[grp] }
    if grp in topology.nodes:
      common.error(
        f"group {grp} is also a node name. I can't deal with that level of confusion",
        common.IncorrectValue,
        'groups')

  '''
  Sanity checks on global group data
  '''

  list_of_modules = modules.list_of_modules(topology)
  group_attr = topology.defaults.attributes.group
  providers = list(topology.defaults.providers.keys())
  for grp,gdata in topology.groups.items():
    must_be_id(parent=None,key=grp,path=f'NOATTR:group name {grp}',module='groups')
    if must_be_dict(topology.groups,grp,'topology.groups',create_empty=True,module='groups') is None:
      continue

    gpath=f'topology.groups.{grp}'
    g_modules = gdata.get('module',[])
    if g_modules:                          # Modules specified in the group -- we know what these nodes will use
      gm_source = 'group'
    else:
      gm_source = 'topology'
      g_modules = topology.get('module',[])

    validate_attributes(
      data=gdata,
      topology=topology,
      data_path=gpath,
      data_name='group',
      attr_list=[ 'group','node' ],
      module='groups',
      modules=g_modules,
      module_source=gm_source,
      extra_attributes=providers)          # Allow provider-specific settings (not checked at the moment)

    if not 'members' in gdata:
      gdata.members = []

    if grp == 'all' and gdata.members:
      common.error('Group "all" should not have explicit members')

    must_be_dict(gdata,'vars',gpath,create_empty=False,module='groups')
    must_be_dict(gdata,'node_data',gpath,create_empty=False,module='groups')
    must_be_list(gdata,'config',gpath,create_empty=False,module='groups')
    must_be_list(gdata,'module',gpath,create_empty=False,module='groups',valid_values=list_of_modules)
    must_be_string(gdata,'device',gpath,module='groups',valid_values=list(topology.defaults.devices))

    if 'node_data' in gdata:                 # Validate node_data attributes (if any)
      validate_attributes(
        data=gdata.node_data,
        topology=topology,
        data_path=f'{gpath}.node_data',
        data_name='node',
        attr_list=[ 'node' ],
        module='groups',
        modules=g_modules,
        module_source=gm_source,
        extra_attributes=providers)          # Allow provider-specific settings (not checked at the moment)

      for k in ('module','device'):          # Check that the 'module' or 'device' attributes are not in node_data
        if k in gdata.node_data:
          common.error(
            f'Cannot use attribute {k} in node_data in group {grp}, set it as a group attribute',
            common.IncorrectValue,
            'groups')

    for k in list(gdata.keys()):             # Then move (validated) group node attributes into node_data
      if k in group_attr:
        continue
      gdata.node_data[k] = gdata[k]
      gdata.pop(k,None)

    if must_be_list(gdata,'members',gpath,create_empty=False,module='groups') is None:
      continue

    for n in gdata.members:
      if not n in topology.nodes and not n in topology.groups:
        common.error('Member %s of group %s is not a valid node or group name' % (n,grp))

'''
Add node-level group settings to global groups
'''
def add_node_level_groups(topology: Box) -> None:
  for name,n in topology.nodes.items():
    if not 'group' in n:
      continue

    must_be_list(n,'group',f'nodes.{name}')

    for grpname in n.group:
      if not grpname in topology.groups:
        topology.groups[grpname] = { 'members': [] }      # Create an empty new group if needed

      if not name in topology.groups[grpname].members:  # Node not yet in the target group
        topology.groups[grpname].members.append(name)   # Add node to the end of the member list

'''
Check recursive group definitions
'''

def check_recursive_chain(topology: Box, chain: list, group: str) -> typing.Optional[list]:
  if not group in topology.groups: # pragma: no cover (if we ever get here we're seriously messed up)
    common.fatal(
      'Internal error: unknown group in check_recursive_chain')
    return None

  chain = chain + [ group ]
  for m in topology.groups[group].members:
    if m in chain:
      chain = chain + [ m ]
      common.error(f'Recursive group definition chain {chain}', common.IncorrectValue, 'groups')
      return chain
    if m in topology.groups:
      if check_recursive_chain(topology,chain,m):
        return chain

  return None

def check_recursive_groups(topology : Box) -> None:
  for gname in topology.groups.keys():
    if check_recursive_chain(topology,[],gname):
      return

def reverse_topsort(topology: Box) -> list:
  group_copy = Box(topology.groups)            # Make a copy of the group dictionary
  sort_list: typing.List[str] = []
  while group_copy:                            # Keep iterating until we got all groups in order
    for g in sorted(group_copy.keys()):        # Iterate over remaining groups
      OK_to_add = True                         # ... sort them by name to have consistent results
      for m in group_copy[g].members:          # Now go check the group members
        if m in group_copy:                    # ... if a member of this group is still in groups we're not yet ready
          OK_to_add = False                    # ... to add it to sorted list
      if OK_to_add:                            # Have all child groups already been removed from the group dictionary?
        sort_list = sort_list + [ g ]          # ... if so, we're good to go, add current group to sorted list
        group_copy.pop(g)                      # ... and remove it from dictionary of remaining groups

  return sort_list

'''
Copy group-level module or device setting into node data
'''
def copy_group_device_module(topology: Box) -> None:
  for grp in reverse_topsort(topology):
    gdata = topology.groups[grp]
    if not 'device' in gdata and not 'module' in gdata:
      continue                                                        # This group is not interesting, move on

    if common.debug_active('groups'):
      print(f'Setting device/module for group {grp}')
    g_members = group_members(topology,grp)
    if not g_members:
      common.error(
        f'Cannot use "module" or "device" attribute on in group {grp} that has no direct or indirect members',
        common.IncorrectValue,
        'groups')
      continue

    for name in g_members:                                            # Iterate over group members
      if not name in topology.nodes:                                  # Member not a node? Move on...
        continue

      ndata = topology.nodes[name]
      if 'device' in gdata and not 'device' in ndata:                 # Copy device from group data to node data
        ndata.device = gdata.device
        if common.debug_active('groups'):
          print(f'... setting {name}.device to {gdata.device}')

      if 'module' in gdata:                                           # Merge group modules with device modules
        ndata.module = ndata.module or []                             # Make sure node.module is a list
        for m in gdata.module:                                        # Now iterate over group modules
          if not m in ndata.module:                                   # ... and add missing modules to nodes
            ndata.module.append(m)

        if common.debug_active('groups'):
          print(f'... adding module {gdata.module} to {name}. Node modules now {ndata.module}')

'''
Copy node data from group into group members
'''
def copy_group_node_data(topology: Box,pfx: str) -> None:
  for grp in reverse_topsort(topology):
    if not grp.startswith(pfx):                                       # Skip groups that don't match the current prefix (ex: BGP autogroups)
      continue
    gdata = topology.groups[grp]
    if not 'node_data' in gdata:                                      # No group data, skip
      continue

    g_members = group_members(topology,grp)                           # Get recursive list of members
    if common.debug_active('groups'):
      print(f'copy node data {grp}: {gdata.node_data}')
    for name in g_members:                                            # Iterate over group members
      if not name in topology.nodes:                                  # Member is not a node, skip it
        continue

      if common.debug_active('groups'):
        print(f'... merging node data with {name}')
      topology.nodes[name] = gdata.node_data + topology.nodes[name]

'''
Export node_data from groups to topology

Used to create module-specific data structures in pre_transform hook before the
node data is populated from groups.node_data

Inputs:
* data structure name ('vrfs' or 'vlans')
* module name ('vrf' or 'vlan') -- used in error messages
* copy_keys: attributes that should be copied to topology-level data structure
* unique_keys: attributes that must be unique (set to copy_keys when empty)
'''
def export_group_node_data(
      topology: Box,
      key: str,
      module: str,
      copy_keys: typing.List[str] = [],
      unique_keys: typing.List[str] = []) -> None:

  if not unique_keys:
    unique_keys = copy_keys
  for gname,gdata in topology.groups.items():
    #
    # Find groups with node_data dictionaries
    if must_be_dict(gdata,f'node_data.{key}',f'groups.{gname}',module=module,create_empty=False):
      for obj_name in list(gdata.node_data[key].keys()):
        if gdata.node_data[key][obj_name] is None:
          gdata.node_data[key][obj_name] = {}
        obj_data =  gdata.node_data[key][obj_name]
        if not obj_name in topology[key] or topology[key][obj_name] is None:
          topology[key][obj_name] = {}
        for attr in unique_keys:
          if attr in topology[key][obj_name] and attr in obj_data and \
             topology[key][obj_name][attr] != obj_data[attr]:                          # Unique key present on both ends and not equal
            common.error(
              f'Cannot redefine {key} attribute {attr} for {key}.{obj_name} from node_data in group {gname}',
              common.IncorrectValue,
              module)
        for attr in copy_keys:
          if attr in obj_data and attr not in topology[key][obj_name]:
            topology[key][obj_name][attr] = obj_data[attr]

#
# create_bgp_autogroups -- create BGP AS groups
#

def create_bgp_autogroups(topology: Box) -> None:
  g_module = topology.get('module',[])                          # Global list of modules, could be invalid
  if not isinstance(g_module,list):                             # Won't deal with incorrectly formatted 'module' attribute here
    g_module = []                                               # ... just assume it's an empty list
  g_bgpas = data.get_global_parameter(topology,'bgp.as')        # Try to get the global BGP AS
  if not data.is_true_int(g_bgpas):                             # ... don't worry if it's not int, someone else will complain
    g_bgpas = 0

  for gname,gdata in topology.groups.items():                   # Sanity check: BGP autogroups should not have static members
    if re.match('as\\d+$',gname):
      if gdata.get('members',None):                             # Well, it's OK to have an empty list of members ;)
        common.error(
          f'BGP AS group {gname} should not have static members',
          common.IncorrectValue,
          'groups')

  for n_name,n_data in topology.nodes.items():                  # Now iterate over nodes
    n_module = n_data.get('module',g_module)                    # Get node or global module (global modules haven't been propagated yet)
    if not isinstance(n_module,list):                           # Node list of modules is insane, someone else will complain
      continue

    if not 'bgp' in n_module:                                   # Looks like this node does not care about BGP
      continue

    n_bgpas = data.get_from_box(n_data,'bgp.as') or g_bgpas     # Get node-level or global BGP AS
    if not n_bgpas:
      continue

    grpname = f"as{n_bgpas}"                                    # BGP auto-group name
    if not 'members' in topology.groups[grpname]:               # Set members of a new group to an empty list
      topology.groups[grpname].members = []                     # ... because we're using Box this also creates an empty group dict

    topology.groups[grpname].members.append(n_name)             # ... and append the node to AS group

#
# init_groups:
#
# * Check and adjust group data structures
# * Check recursive groups
# * Add nodes to groups based on node 'group' attribute
#
# Please note that check_group_data_structure creates 'groups' element if needed
# and 'adjust_groups' deletes it if there are no groups in the topology.
#
def init_groups(topology: Box) -> None:
  check_group_data_structure(topology)
  add_node_level_groups(topology)
  common.exit_on_error()

  check_recursive_groups(topology)
  common.exit_on_error()

  copy_group_device_module(topology)
  copy_group_node_data(topology,'')                 # Copy all group data into nodes (potentially setting bgp.as)
  bgp.process_as_list(topology)
  create_bgp_autogroups(topology)                   # Create AS-based groups
  copy_group_node_data(topology,'as')               # And add group data from 'asxxxx' into nodes
  common.exit_on_error()
  if not topology.groups:
    del topology['groups']

'''
Copy custom config templates from groups into nodes
'''
def node_config_templates(topology: Box) -> None:
  if not 'groups' in topology:
    return

  '''
  Phase 1 - merge group config templates into nodes

  Traverse groups from more-specific to less specific, pushing config templates in front of
  the accumulated config list.

  End result: config templates sorted from less specific through more specific groups, ending
  with node templates.
  '''

  for group_name in reverse_topsort(topology):
    if not 'config' in topology.groups[group_name]:
      continue

    must_be_list(topology.groups[group_name],'config',f'groups.{group_name}')
    g_members = group_members(topology,group_name)
    for name,ndata in topology.nodes.items():
      if name in g_members or group_name == 'all':
        if not must_be_list(ndata,'config',f'nodes.{name}') is None:
          ndata.config = topology.groups[group_name].config + ndata.config

  '''
  Phase 2 - cleanup

  * Remove 'config' attributes from groups (just in case, they are no longer needed anyway)
  * Process node 'config' attributes to honor 'removal of prior templates' requests
  '''

  for group_name in topology.groups:
    topology.groups[group_name].pop('config',None)

  for name,node in topology.nodes.items():
    if not 'config' in node:
      continue

    config_list: typing.List[str] = []
    for c in node.config:
      if c == '-':                 # Remove all prior templates
        config_list = []
      elif c[0] == '-':
        config_list = [ t for t in config_list if t != c [1:] ]
      else:
        config_list = config_list + [ c ]

    if config_list:
      node.config = config_list
    else:
      node.pop('config',None)



#
# Dynamic virtualization provider framework
#
# Individual virtualization providers are defined in modules within this directory inheriting
# Provider class and replacing or augmenting its methods (most commonly, transform)
#

import platform
import subprocess
import os
import typing

# Related modules
from box import Box

from .. import common
from ..callback import Callback
from ..augment import devices
from ..data import get_from_box,get_box

class _Provider(Callback):
  def __init__(self, provider: str, data: Box) -> None:
    self.provider = provider
    if 'template' in data:
      self._default_template_name = data.template

  @classmethod
  def load(self, provider: str, data: Box) -> '_Provider':
    module_name = __name__+"."+provider
    obj = self.find_class(module_name)
    if obj:
      return obj(provider,data)
    else:
      return _Provider(provider,data)

  def get_template_path(self) -> str:
    return 'templates/provider/' + self.provider

  def get_full_template_path(self) -> str:
    return str(common.get_moddir()) + '/' + self.get_template_path()

  def find_extra_template(self, node: Box, fname: str) -> typing.Optional[str]:
    return common.find_file(fname+'.j2',[ f'./{node.device}','.',f'{ self.get_full_template_path() }/{node.device}'])

  def get_output_name(self, fname: typing.Optional[str], topology: Box) -> str:
    if fname:
      return fname

    fname = topology.defaults.providers[self.provider].config
    if fname:
      return fname

    return "Vagrantfile"

  _default_template_name = "Vagrantfile.j2"

  def get_root_template(self) -> str:
    return self._default_template_name

  def node_image_version(self, topology: Box) -> None:
    for name,n in topology.nodes.items():
      if '.' in n.box:
        image_spec = n.box.split(':')
        n.box = image_spec[0]
        if len(image_spec) > 1:
          n.box_version = image_spec[1]

  def transform_node_images(self, topology: Box) -> None:
    pass

  def transform(self, topology: Box) -> None:
    self.transform_node_images(topology)
    if "processor" in topology.defaults:
      return
    else:
      processor_name = ""
      if platform.system() == "Windows":
        processor_name = platform.processor()
      elif platform.system() == "Darwin":
        processor_name = "intel"  # Assume Intel for MacOS
      elif platform.system() == "Linux":
        processor_name = str(subprocess.check_output("cat /proc/cpuinfo", shell=True).splitlines()[1].split()[2])
      topology.defaults.processor = processor_name

  def create_extra_files_mappings(
      self,
      node: Box,
      topology: Box,
      inkey: str = 'config_templates',
      outkey: str = 'binds') -> None:

    mappings = get_from_box(node,f'{self.provider}.{inkey}')
    if mappings:
      cur_binds = get_from_box(node,f'{self.provider}.{outkey}') or {}
      for file,mapping in mappings.items():
        if mapping in cur_binds.values():
          continue
        if not isinstance(mapping,str):
          common.error(
            f"Malformed extra file mapping for {self.provider}.{inkey}.{file} on node {node.name} -- should be string",
            common.IncorrectType,
            self.provider)
          continue

        if not self.find_extra_template(node,file):
          common.error(
            f"Cannot find template {file}.j2 for extra file {self.provider}.{inkey}.{file} on node {node.name}",
            common.IncorrectValue,
            self.provider)
          continue

        out_folder = f"{self.provider}_files/{node.name}"
        node[self.provider][outkey][f"{out_folder}/{file}"] = mapping

  def create_extra_files(
      self,
      node: Box,
      topology: Box,
      inkey: str = 'config_templates',
      outkey: str = 'binds') -> None:

    binds = get_from_box(node,f'{self.provider}.{outkey}')
    if not binds:
      return

    sys_folder = str(common.get_moddir())+"/"
    out_folder = f"{self.provider}_files/{node.name}"

    for file,mapping in binds.items():
      file_name = file.replace(out_folder+"/","")
      template_name = self.find_extra_template(node,file_name)
      if template_name:
        node_data = node + { 'hostvars': topology.nodes }
        common.write_template(
          in_folder=os.path.dirname(template_name),
          j2=os.path.basename(template_name),
          data=node_data.to_dict(),
          out_folder=out_folder, filename=file_name)
        print( f"Created {out_folder}/{file_name} from {template_name.replace(sys_folder,'')}, mapped to {node.name}:{mapping}" )

  def create(self, topology: Box, fname: typing.Optional[str]) -> None:
    self.transform(topology)
    fname = self.get_output_name(fname,topology)
    output = common.open_output_file(fname)
#    print(topology.nodes.to_yaml())
    output.write(common.template(self.get_root_template(),topology.to_dict(),self.get_template_path(),self.provider))
    if fname != '-':
      common.close_output_file(output)
      print("Created provider configuration file: %s" % fname)
      self.post_configuration_create(topology)
    else:
      output.write("\n")

  def post_start_lab(self, topology: Box) -> None:
    pass

  def pre_start_lab(self, topology: Box) -> None:
    pass

  def pre_stop_lab(self, topology: Box) -> None:
    pass

  def post_stop_lab(self, topology: Box) -> None:
    pass

  def post_configuration_create(self, topology: Box) -> None:
    pass

  """
  Generic provider pre-transform processing: Mark multi-provider links
  """
  def pre_transform(self,topology : Box) -> None:
    if not 'links' in topology:
      return

    for l in topology.links:
      for intf in l.interfaces:
        node = topology.nodes[intf.node]
        if not 'provider' in node:
          continue

        l[topology.provider].provider[node.provider] = True

  """
  Generic provider pre-output transform: remove loopback links
  """
  def pre_output_transform(self, topology: Box) -> None:
    if not 'links' in topology:
      return

    topology.links = [ link for link in topology.links if link.type != 'loopback' ]

"""
Get a pointer to provider module. Cached in topology._Providers
"""
def get_provider_module(topology: Box, pname: str) -> _Provider:
  if not pname in topology._Providers:
    topology._Providers[pname] = _Provider.load(pname,topology.defaults.providers[pname])

  return topology._Providers[pname]

"""
Execute a topology-wide provider hook
"""
def execute(hook: str, topology: Box) -> None:
  p_module = get_provider_module(topology,topology.provider)
  p_module.call(hook,topology)

"""
Execute a node-level provider hook
"""
def execute_node(hook: str, node: Box, topology: Box) -> None:
  node_provider = devices.get_provider(node,topology.defaults)
  p_module = get_provider_module(topology,node_provider)
  p_module.call(hook,node,topology)

"""
Mark all nodes and links with relevant provider(s)
"""
def mark_providers(topology: Box) -> None:
  for n in topology.nodes.values():                 # Set 'provider' attribute on all nodes
    if 'provider' in n:
      continue

    n.provider = topology.provider

  for l in topology.links:                          # Set 'providers' attribute on all links
    for intf in l.interfaces:
      node = topology.nodes[intf.node]
      l.provider[node.provider] = True

"""
Select a subset of the topology -- links and nodes relevant to the current provider
"""
def select_topology(topology: Box, provider: str) -> Box:
  topology = get_box(topology)                      # Create a copy of the topology
  for n in list(topology.nodes.keys()):             # Remove all nodes not belonging to the current provider
    if topology.nodes[n].provider != provider:
      topology.nodes.pop(n,None)

  topology.links = [ l for l in topology.links if provider in l.provider ]      # Retain only the links used by current provider
  return topology
